@@ -2,7 +2,7 @@
  *			GPAC - Multimedia Framework C SDK
  *
  *			Authors: Jean Le Feuvre
- *			Copyright (c) Telecom ParisTech 2017-2020
+ *			Copyright (c) Telecom ParisTech 2017-2022
  *					All rights reserved
  *
  *  This file is part of GPAC / generic FILE output filter
@@ -28,13 +28,22 @@
 #include <gpac/constants.h>
 #include <gpac/xml.h>
 
+enum
+{
+	FOUT_CAT_NONE = 0,
+	FOUT_CAT_AUTO,
+	FOUT_CAT_ALL
+};
+
 typedef struct
 {
 	//options
 	Double start, speed;
 	char *dst, *mime, *ext;
-	Bool append, dynext, cat, ow, redund;
+	Bool append, dynext, ow, redund, noinitraw;
+	u32 cat;
 	u32 mvbk;
+	s32 max_cache_segs;
 
 	//only one input pid
 	GF_FilterPid *pid;
@@ -42,18 +51,23 @@ typedef struct
 	FILE *file;
 	Bool is_std;
 	u64 nb_write;
-
+	Bool use_templates;
 	GF_FilterCapability in_caps[2];
 	char szExt[10];
 	char szFileName[GF_MAX_PATH];
 
 	Bool patch_blocks;
 	Bool is_null;
-	GF_Err is_error;
+	GF_Err error;
 	u32 dash_mode;
 	u64 offset_at_seg_start;
 	const char *original_url;
 	GF_FileIO *gfio_ref;
+
+	FILE *hls_chunk;
+
+	u32 max_segs;
+	GF_List *past_files;
 } GF_FileOutCtx;
 
 #ifdef WIN32
@@ -61,11 +75,20 @@ typedef struct
 #include <fcntl.h>
 #endif //WIN32
 
+static void fileout_close_hls_chunk(GF_FileOutCtx *ctx, Bool final_flush)
+{
+	if (!ctx->hls_chunk) return;
+	gf_fclose(ctx->hls_chunk);
+	ctx->hls_chunk = NULL;
+}
+
 static GF_Err fileout_open_close(GF_FileOutCtx *ctx, const char *filename, const char *ext, u32 file_idx, Bool explicit_overwrite, char *file_suffix)
 {
 	if (ctx->file && !ctx->is_std) {
 		GF_LOG(GF_LOG_INFO, GF_LOG_MMIO, ("[FileOut] closing output file %s\n", ctx->szFileName));
 		gf_fclose(ctx->file);
+
+		fileout_close_hls_chunk(ctx, GF_FALSE);
 	}
 	ctx->file = NULL;
 
@@ -83,7 +106,7 @@ static GF_Err fileout_open_close(GF_FileOutCtx *ctx, const char *filename, const
 #endif
 
 	} else {
-		char szName[GF_MAX_PATH], szFinalName[GF_MAX_PATH];
+		char szFinalName[GF_MAX_PATH];
 		Bool append = ctx->append;
 		const char *url = filename;
 
@@ -93,18 +116,38 @@ static GF_Err fileout_open_close(GF_FileOutCtx *ctx, const char *filename, const
 		if (ctx->dynext) {
 			const char *has_ext = gf_file_ext_start(url);
 
-			strcpy(szName, url);
+			strcpy(szFinalName, url);
 			if (!has_ext && ext) {
-				strcat(szName, ".");
-				strcat(szName, ext);
+				strcat(szFinalName, ".");
+				strcat(szFinalName, ext);
 			}
 		} else {
-			strcpy(szName, url);
+			strcpy(szFinalName, url);
+		}
+
+		if (ctx->use_templates) {
+			GF_Err e;
+			char szName[GF_MAX_PATH];
+			assert(ctx->dst);
+			if (!strcmp(filename, ctx->dst)) {
+				strcpy(szName, szFinalName);
+				e = gf_filter_pid_resolve_file_template(ctx->pid, szName, szFinalName, file_idx, file_suffix);
+			} else {
+				char szFileName[GF_MAX_PATH];
+				strcpy(szFileName, szFinalName);
+				strcpy(szName, ctx->dst);
+				e = gf_filter_pid_resolve_file_template_ex(ctx->pid, szName, szFinalName, file_idx, file_suffix, szFileName);
+			}
+			if (e) {
+				return ctx->error = e;
+			}
 		}
-		gf_filter_pid_resolve_file_template(ctx->pid, szName, szFinalName, file_idx, file_suffix);
 
 		if (!gf_file_exists(szFinalName)) append = GF_FALSE;
 
+		if (!strcmp(szFinalName, ctx->szFileName) && (ctx->cat==FOUT_CAT_AUTO))
+			append = GF_TRUE;
+
 		if (!ctx->ow && gf_file_exists(szFinalName) && !append) {
 			char szRes[21];
 			s32 res;
@@ -112,23 +155,23 @@ static GF_Err fileout_open_close(GF_FileOutCtx *ctx, const char *filename, const
 			fprintf(stderr, "File %s already exist - override (y/n/a) ?:", szFinalName);
 			res = scanf("%20s", szRes);
 			if (!res || (szRes[0] == 'n') || (szRes[0] == 'N')) {
-				return ctx->is_error = GF_IO_ERR;
+				return ctx->error = GF_IO_ERR;
 			}
 			if ((szRes[0] == 'a') || (szRes[0] == 'A')) ctx->ow = GF_TRUE;
 		}
 
 		GF_LOG(GF_LOG_INFO, GF_LOG_MMIO, ("[FileOut] opening output file %s\n", szFinalName));
-		ctx->file = gf_fopen_ex(szFinalName, ctx->original_url, append ? "a+b" : "w+b");
+		ctx->file = gf_fopen_ex(szFinalName, ctx->original_url, append ? "a+b" : "w+b", GF_FALSE);
 
-		if (!strcmp(szFinalName, ctx->szFileName) && !ctx->append && ctx->nb_write && !explicit_overwrite) {
-			GF_LOG(GF_LOG_WARNING, GF_LOG_MMIO, ("[FileOut] re-opening in write mode output file %s, content overwrite\n", szFinalName));
+		if (!strcmp(szFinalName, ctx->szFileName) && !append && ctx->nb_write && !explicit_overwrite) {
+			GF_LOG(GF_LOG_WARNING, GF_LOG_MMIO, ("[FileOut] re-opening in write mode output file %s, content overwrite (use `cat` option to enable append)\n", szFinalName));
 		}
 		strcpy(ctx->szFileName, szFinalName);
 	}
 	ctx->nb_write = 0;
 	if (!ctx->file) {
 		GF_LOG(GF_LOG_ERROR, GF_LOG_MMIO, ("[FileOut] cannot open output file %s\n", ctx->szFileName));
-		return ctx->is_error = GF_IO_ERR;;
+		return ctx->error = GF_IO_ERR;;
 	}
 
 	return GF_OK;
@@ -136,21 +179,46 @@ static GF_Err fileout_open_close(GF_FileOutCtx *ctx, const char *filename, const
 
 static void fileout_setup_file(GF_FileOutCtx *ctx, Bool explicit_overwrite)
 {
+	const char *dst = ctx->dst;
 	const GF_PropertyValue *p, *ext;
+
 	p = gf_filter_pid_get_property(ctx->pid, GF_PROP_PID_OUTPATH);
 	ext = gf_filter_pid_get_property(ctx->pid, GF_PROP_PID_FILE_EXT);
 
 	if (p && p->value.string) {
 		fileout_open_close(ctx, p->value.string, (ext && ctx->dynext) ? ext->value.string : NULL, 0, explicit_overwrite, NULL);
-	} else if (ctx->dynext) {
+		return;
+	}
+	if (!dst) {
+		p = gf_filter_pid_get_property(ctx->pid, GF_PROP_PID_FILEPATH);
+		if (p && p->value.string) {
+			dst = p->value.string;
+			char *sep = strstr(dst, "://");
+			if (sep) {
+				dst = strchr(sep+3, '/');
+				if (!dst) return;
+			} else {
+				if (!strncmp(dst, "./", 2)) dst+= 2;
+				else if (!strncmp(dst, ".\\", 2)) dst+= 2;
+				else if (!strncmp(dst, "../", 3)) dst+= 3;
+				else if (!strncmp(dst, "..\\", 3)) dst+= 3;
+			}
+		}
+	}
+	if (ctx->dynext) {
 		p = gf_filter_pid_get_property(ctx->pid, GF_PROP_PCK_FILENUM);
 		if (!p) {
 			if (ext && ext->value.string) {
-				fileout_open_close(ctx, ctx->dst, ext->value.string, 0, explicit_overwrite, NULL);
+				fileout_open_close(ctx, dst, ext->value.string, 0, explicit_overwrite, NULL);
 			}
 		}
-	} else {
+	} else if (ctx->dst) {
 		fileout_open_close(ctx, ctx->dst, NULL, 0, explicit_overwrite, NULL);
+	} else {
+		p = gf_filter_pid_get_property(ctx->pid, GF_PROP_PID_FILEPATH);
+		if (!p) p = gf_filter_pid_get_property(ctx->pid, GF_PROP_PID_URL);
+		if (p && p->value.string)
+			fileout_open_close(ctx, p->value.string, NULL, 0, explicit_overwrite, NULL);
 	}
 }
 static GF_Err fileout_configure_pid(GF_Filter *filter, GF_FilterPid *pid, Bool is_remove)
@@ -176,12 +244,26 @@ static GF_Err fileout_configure_pid(GF_Filter *filter, GF_FilterPid *pid, Bool i
 
 	p = gf_filter_pid_get_property(pid, GF_PROP_PID_DASH_MODE);
 	if (p && p->value.uint) ctx->dash_mode = 1;
+
+	ctx->max_segs = 0;
+	p = gf_filter_pid_get_property(pid, GF_PROP_PID_TIMESHIFT_SEGS);
+	if (ctx->max_cache_segs<0) {
+		ctx->max_segs = (u32) -ctx->max_cache_segs;
+	} else if (ctx->max_cache_segs>0) {
+		ctx->max_segs = (u32) ctx->max_cache_segs;
+		if (p && (p->value.uint > (u32) ctx->max_cache_segs))
+			ctx->max_segs = p->value.uint;
+	}
+	if (ctx->max_segs && !ctx->past_files)
+		ctx->past_files = gf_list_new();
+
+	ctx->error = GF_OK;
 	return GF_OK;
 }
 
 static GF_Err fileout_initialize(GF_Filter *filter)
 {
-	char *ext=NULL;
+	char *ext=NULL, *sep;
 	const char *dst;
 	GF_FileOutCtx *ctx = (GF_FileOutCtx *) gf_filter_get_udta(filter);
 
@@ -218,6 +300,12 @@ static GF_Err fileout_initialize(GF_Filter *filter)
 		dst = ctx->dst;
 	}
 
+	sep = dst ? strchr(dst, '$') : NULL;
+	if (sep) {
+		sep = strchr(sep+1, '$');
+		if (sep) ctx->use_templates = GF_TRUE;
+	}
+
 	if (ctx->dynext) return GF_OK;
 
 	if (ctx->ext) ext = ctx->ext;
@@ -257,9 +345,20 @@ static void fileout_finalize(GF_Filter *filter)
 {
 	GF_Err e;
 	GF_FileOutCtx *ctx = (GF_FileOutCtx *) gf_filter_get_udta(filter);
+
+	fileout_close_hls_chunk(ctx, GF_TRUE);
+
 	fileout_open_close(ctx, NULL, NULL, 0, GF_FALSE, NULL);
 	if (ctx->gfio_ref)
 		gf_fileio_open_url((GF_FileIO *)ctx->gfio_ref, NULL, "unref", &e);
+
+	if (ctx->past_files) {
+		while (gf_list_count(ctx->past_files)) {
+			char *url = gf_list_pop_back(ctx->past_files);
+			gf_free(url);
+		}
+		gf_list_del(ctx->past_files);
+	}
 }
 
 static GF_Err fileout_process(GF_Filter *filter)
@@ -271,14 +370,8 @@ static GF_Err fileout_process(GF_Filter *filter)
 	u32 pck_size, nb_write;
 	GF_FileOutCtx *ctx = (GF_FileOutCtx *) gf_filter_get_udta(filter);
 
-	if (ctx->is_error) {
-		GF_Err e = ctx->is_error;
-		if (e != GF_EOS) {
-			gf_filter_pid_set_discard(ctx->pid, GF_TRUE);
-			ctx->is_error = GF_EOS;
-		}
-		return e;
-	}
+	if (ctx->error)
+		return ctx->error;
 
 	pck = gf_filter_pid_get_packet(ctx->pid);
 	if (!pck) {
@@ -295,13 +388,13 @@ static GF_Err fileout_process(GF_Filter *filter)
 				evt.seg_size.seg_url = NULL;
 
 				if (ctx->dash_mode==1) {
-					evt.seg_size.is_init = GF_TRUE;
+					evt.seg_size.is_init = 1;
 					ctx->dash_mode = 2;
 					evt.seg_size.media_range_start = 0;
 					evt.seg_size.media_range_end = 0;
 					gf_filter_pid_send_event(ctx->pid, &evt);
 				} else {
-					evt.seg_size.is_init = GF_FALSE;
+					evt.seg_size.is_init = 0;
 					evt.seg_size.media_range_start = ctx->offset_at_seg_start;
 					evt.seg_size.media_range_end = gf_ftell(ctx->file)-1;
 					gf_filter_pid_send_event(ctx->pid, &evt);
@@ -314,7 +407,7 @@ static GF_Err fileout_process(GF_Filter *filter)
 	}
 
 	gf_filter_pck_get_framing(pck, &start, &end);
-	if (!ctx->redund) {
+	if (!ctx->redund && !ctx->dash_mode) {
 		u32 dep_flags = gf_filter_pck_get_dependency_flags(pck);
 		//redundant packet, do not store
 		if ((dep_flags & 0x3) == 1) {
@@ -347,7 +440,7 @@ static GF_Err fileout_process(GF_Filter *filter)
 		return fileout_process(filter);
 	}
 
-	if (ctx->file && start && ctx->cat)
+	if (ctx->file && start && (ctx->cat==FOUT_CAT_ALL))
 		start = GF_FALSE;
 
 	if (ctx->dash_mode) {
@@ -359,16 +452,16 @@ static GF_Err fileout_process(GF_Filter *filter)
 			evt.seg_size.seg_url = NULL;
 
 			if (ctx->dash_mode==1) {
-				evt.seg_size.is_init = GF_TRUE;
+				evt.seg_size.is_init = 1;
 				ctx->dash_mode = 2;
 				evt.seg_size.media_range_start = 0;
 				evt.seg_size.media_range_end = 0;
 				gf_filter_pid_send_event(ctx->pid, &evt);
 			} else {
-				evt.seg_size.is_init = GF_FALSE;
+				evt.seg_size.is_init = 0;
 				evt.seg_size.media_range_start = ctx->offset_at_seg_start;
 				evt.seg_size.media_range_end = gf_ftell(ctx->file)-1;
-				ctx->offset_at_seg_start = evt.seg_size.media_range_end;
+				ctx->offset_at_seg_start = evt.seg_size.media_range_end+1;
 				gf_filter_pid_send_event(ctx->pid, &evt);
 			}
 			if ( gf_filter_pck_get_property(pck, GF_PROP_PCK_FILENAME))
@@ -401,9 +494,29 @@ static GF_Err fileout_process(GF_Filter *filter)
 
 		if (name) {
 			fileout_open_close(ctx, name, ext ? ext->value.string : NULL, fnum ? fnum->value.uint : 0, explicit_overwrite, fsuf ? fsuf->value.string : NULL);
-		} else if (!ctx->file) {
+		} else if (!ctx->file && !ctx->noinitraw) {
 			fileout_setup_file(ctx, explicit_overwrite);
 		}
+
+		if (ctx->max_segs) {
+			while (gf_list_count(ctx->past_files)>ctx->max_segs) {
+				char *url = gf_list_pop_front(ctx->past_files);
+				gf_file_delete(url);
+				gf_free(url);
+			}
+			p = gf_filter_pck_get_property(pck, GF_PROP_PCK_INIT);
+			if (!p || !p->value.boolean) {
+				gf_list_add(ctx->past_files, gf_strdup(ctx->szFileName));
+			}
+		}
+	}
+
+	p = gf_filter_pck_get_property(pck, GF_PROP_PCK_HLS_FRAG_NUM);
+	if (p) {
+		char szHLSChunk[GF_MAX_PATH+21];
+		snprintf(szHLSChunk, GF_MAX_PATH+20, "%s.%d", ctx->szFileName, p->value.uint);
+		if (ctx->hls_chunk) gf_fclose(ctx->hls_chunk);
+		ctx->hls_chunk = gf_fopen_ex(szHLSChunk, ctx->original_url, "w+b", GF_FALSE);
 	}
 
 	pck_data = gf_filter_pck_get_data(pck, &pck_size);
@@ -440,7 +553,7 @@ static GF_Err fileout_process(GF_Filter *filter)
 						pos = cur_w;
 						block = gf_malloc(ctx->mvbk);
 						if (!block) {
-							GF_LOG(GF_LOG_ERROR, GF_LOG_MMIO, ("[FileOut] unable to allocate blockof %d bytes\n", ctx->mvbk));
+							GF_LOG(GF_LOG_ERROR, GF_LOG_MMIO, ("[FileOut] unable to allocate block of %d bytes\n", ctx->mvbk));
 						} else {
 							while (cur_r > bo) {
 								u32 move_bytes = ctx->mvbk;
@@ -481,6 +594,13 @@ static GF_Err fileout_process(GF_Filter *filter)
 					GF_LOG(GF_LOG_ERROR, GF_LOG_MMIO, ("[FileOut] Write error, wrote %d bytes but had %d to write\n", nb_write, pck_size));
 				}
 				ctx->nb_write += nb_write;
+
+				if (ctx->hls_chunk) {
+					nb_write = (u32) gf_fwrite(pck_data, pck_size, ctx->hls_chunk);
+					if (nb_write!=pck_size) {
+						GF_LOG(GF_LOG_ERROR, GF_LOG_MMIO, ("[FileOut] Write error, wrote %d bytes but had %d to write\n", nb_write, pck_size));
+					}
+				}
 			}
 		} else if (hwf) {
 			u32 w, h, stride, stride_uv, pf;
@@ -547,6 +667,7 @@ static Bool fileout_process_event(GF_Filter *filter, const GF_FilterEvent *evt)
 		if (ctx->is_null) {
 			GF_LOG(GF_LOG_INFO, GF_LOG_MMIO, ("[FileOut] null delete (file name was %s)\n", evt->file_del.url));
 		} else {
+			GF_LOG(GF_LOG_INFO, GF_LOG_MMIO, ("[FileOut] delete file %s\n", evt->file_del.url));
 			gf_file_delete(evt->file_del.url);
 		}
 		return GF_TRUE;
@@ -573,18 +694,23 @@ static GF_FilterProbeScore fileout_probe_url(const char *url, const char *mime)
 
 static const GF_FilterArgs FileOutArgs[] =
 {
-	{ OFFS(dst), "location of destination file - see filter help ", GF_PROP_NAME, NULL, NULL, 0},
+	{ OFFS(dst), "location of destination file", GF_PROP_NAME, NULL, NULL, 0},
 	{ OFFS(append), "open in append mode", GF_PROP_BOOL, "false", NULL, GF_FS_ARG_HINT_ADVANCED},
 	{ OFFS(dynext), "indicate the file extension is set by filter chain, not dst", GF_PROP_BOOL, "false", NULL, GF_FS_ARG_HINT_ADVANCED},
-	{ OFFS(start), "set playback start offset. Negative value means percent of media dur with -1 <=> dur", GF_PROP_DOUBLE, "0.0", NULL, 0},
-	{ OFFS(speed), "set playback speed when vsync is on. If speed is negative and start is 0, start is set to -1", GF_PROP_DOUBLE, "1.0", NULL, 0},
+	{ OFFS(start), "set playback start offset. A negative value means percent of media duration with -1 equal to duration", GF_PROP_DOUBLE, "0.0", NULL, 0},
+	{ OFFS(speed), "set playback speed when vsync is on. If negative and start is 0, start is set to -1", GF_PROP_DOUBLE, "1.0", NULL, 0},
 	{ OFFS(ext), "set extension for graph resolution, regardless of file extension", GF_PROP_NAME, NULL, NULL, GF_FS_ARG_HINT_ADVANCED},
 	{ OFFS(mime), "set mime type for graph resolution", GF_PROP_NAME, NULL, NULL, GF_FS_ARG_HINT_EXPERT},
-	{ OFFS(cat), "cat each file of input pid rather than creating one file per filename", GF_PROP_BOOL, "false", NULL, GF_FS_ARG_HINT_ADVANCED},
+	{ OFFS(cat), "cat each file of input PID rather than creating one file per filename\n"
+			"- none: never cat files\n"
+			"- auto: only cat if files have same names\n"
+			"- all: always cat regardless of file names"
+	, GF_PROP_UINT, "none", "none|auto|all", GF_FS_ARG_HINT_ADVANCED},
 	{ OFFS(ow), "overwrite output if existing", GF_PROP_BOOL, "true", NULL, 0},
 	{ OFFS(mvbk), "block size used when moving parts of the file around in patch mode", GF_PROP_UINT, "8192", NULL, 0},
 	{ OFFS(redund), "keep redundant packet in output file", GF_PROP_BOOL, "false", NULL, 0},
-
+	{ OFFS(noinitraw), "do not produce initial segment", GF_PROP_BOOL, "false", NULL, GF_FS_ARG_HINT_HIDE},
+	{ OFFS(max_cache_segs), "maximum number of segments cached per HAS quality when recording live sessions (0 means no limit)", GF_PROP_SINT, "0", NULL, GF_FS_ARG_HINT_EXPERT},
 	{0}
 };
 
@@ -598,14 +724,29 @@ static const GF_FilterCapability FileOutCaps[] =
 GF_FilterRegister FileOutRegister = {
 	.name = "fout",
 	GF_FS_SET_DESCRIPTION("File output")
-	GF_FS_SET_HELP("The file output filter is used to write output to disk, and does not produce any output PID.\n"
-		"It can work as a null sink when its destination is `null`, dropping all input packets. In this case it accepts ANY type of input pid, not just file ones.\n"
-		"In regular mode, the filter only accept pid of type file. It will dump to file incomming packets (stream type file), starting a new file for each packet having a __frame_start__ flag set, unless operating in [-cat]() mode.\n"
-		"The ouput file name can use gpac templating mechanism, see `gpac -h doc`."
+	GF_FS_SET_HELP("This filter is used to write data to disk, and does not produce any output PID.\n"
+		"In regular mode, the filter only accept PID of type file. It will dump to file incoming packets (stream type file), starting a new file for each packet having a __frame_start__ flag set, unless operating in [-cat]() mode.\n"
+		"If the output file name is `std` or `stdout`, writes to stdout.\n"
+		"The output file name can use gpac templating mechanism, see `gpac -h doc`."
 		"The filter watches the property `FileNumber` on incoming packets to create new files.\n"
+		"\n"
+		"# Discard sink mode\n"
+		"When the destination is `null`, the filter is a sink dropping all input packets.\n"
+		"In this case it accepts ANY type of input PID, not just file ones.\n"
+		"\n"
+		"# HTTP streaming recording\n"
+		"When recording a DASH or HLS session, the number of segments to keep per quality can be set using [-max_cache_segs]().\n"
+		"- value `0`  keeps everything (default behaviour)\n"
+		"- a negative value `N` will keep `-N` files regardless of the time-shift buffer value\n"
+		"- a positive value `N` will keep `MAX(N, time-shift buffer)` files\n"
+		"\n"
+		"EX gpac -i LIVE_MPD dashin:forward=file -o rec/$File$:max_cache_segs=3\n"
+		"This will force keeping a maximum of 3 media segments while recording the DASH session.\n"
+		""
 	)
 	.private_size = sizeof(GF_FileOutCtx),
 	.args = FileOutArgs,
+	.flags = GF_FS_REG_FORCE_REMUX,
 	SETCAPS(FileOutCaps),
 	.probe_url = fileout_probe_url,
 	.initialize = fileout_initialize,
